@@ -23,25 +23,32 @@
 
 #include <csync.h>
 
+#include <owncloudlib.h>
+
 namespace OCC {
 
 class SyncFileItem;
 class SyncJournalFileRecord;
-typedef QSharedPointer<SyncFileItem> SyncFileItemPtr;
+using SyncFileItemPtr = QSharedPointer<SyncFileItem>;
 
 /**
  * @brief The SyncFileItem class
  * @ingroup libsync
  */
-class SyncFileItem
+class OWNCLOUDSYNC_EXPORT SyncFileItem
 {
+    Q_GADGET
 public:
     enum Direction {
         None = 0,
         Up,
         Down
     };
+    Q_ENUM(Direction)
 
+    using EncryptionStatus = EncryptionStatusEnums::ItemEncryptionStatus;
+
+    // Note: the order of these statuses is used for ordering in the SortedActivityListModel
     enum Status { // stored in 4 bits
         NoStatus,
 
@@ -49,8 +56,6 @@ public:
         NormalError, ///< Error attached to a particular file
         SoftError, ///< More like an information
 
-        Success, ///< The file was properly synced
-
         /** Marks a conflict, old or new.
          *
          * With instruction:IGNORE: detected an old unresolved old conflict
@@ -62,6 +67,22 @@ public:
         FileLocked, ///< The file is locked
         Restoration, ///< The file was restored because what should have been done was not allowed
 
+        /**
+         * The filename is invalid on this platform and could not created.
+         */
+        FileNameInvalid,
+
+        /**
+         * The filename contains invalid characters and can not be uploaded to the server
+         */
+        FileNameInvalidOnServer,
+
+        /**
+         * There is a file name clash (e.g. attempting to download test.txt when TEST.TXT already exists
+         * on a platform where the filesystem is case-insensitive
+         */
+        FileNameClash,
+
         /** For errors that should only appear in the error view.
          *
          * Some errors also produce a summary message. Usually displaying that message is
@@ -80,10 +101,28 @@ public:
          *
          * A SoftError caused by blacklisting.
          */
-        BlacklistedError
+        BlacklistedError,
+
+        Success, ///< The file was properly synced
+    };
+    Q_ENUM(Status)
+
+    enum class LockStatus {
+        UnlockedItem = 0,
+        LockedItem = 1,
     };
 
-    SyncJournalFileRecord toSyncJournalFileRecordWithInode(const QString &localFileName);
+    Q_ENUM(LockStatus)
+
+    enum class LockOwnerType : int{
+        UserLock = 0,
+        AppLock = 1,
+        TokenLock = 2,
+    };
+
+    Q_ENUM(LockOwnerType)
+
+    [[nodiscard]] SyncJournalFileRecord toSyncJournalFileRecordWithInode(const QString &localFileName) const;
 
     /** Creates a basic SyncFileItem from a DB record
      *
@@ -92,6 +131,10 @@ public:
      */
     static SyncFileItemPtr fromSyncJournalFileRecord(const SyncJournalFileRecord &rec);
 
+    /** Creates a basic SyncFileItem from remote properties
+     */
+    [[nodiscard]] static SyncFileItemPtr fromProperties(const QString &filePath, const QMap<QString, QString> &properties);
+
 
     SyncFileItem()
         : _type(ItemTypeSkip)
@@ -101,6 +144,7 @@ public:
         , _errorMayBeBlacklisted(false)
         , _status(NoStatus)
         , _isRestoration(false)
+        , _isSelectiveSync(false)
     {
     }
 
@@ -143,7 +187,7 @@ public:
         return data1[prefixL] < data2[prefixL];
     }
 
-    QString destination() const
+    [[nodiscard]] QString destination() const
     {
         if (!_renameTarget.isEmpty()) {
             return _renameTarget;
@@ -151,12 +195,12 @@ public:
         return _file;
     }
 
-    bool isEmpty() const
+    [[nodiscard]] bool isEmpty() const
     {
         return _file.isEmpty();
     }
 
-    bool isDirectory() const
+    [[nodiscard]] bool isDirectory() const
     {
         return _type == ItemTypeDirectory;
     }
@@ -164,7 +208,7 @@ public:
     /**
      * True if the item had any kind of error.
      */
-    bool hasErrorStatus() const
+    [[nodiscard]] bool hasErrorStatus() const
     {
         return _status == SyncFileItem::SoftError
             || _status == SyncFileItem::NormalError
@@ -175,7 +219,7 @@ public:
     /**
      * Whether this item should appear on the issues tab.
      */
-    bool showInIssuesTab() const
+    [[nodiscard]] bool showInIssuesTab() const
     {
         return hasErrorStatus() || _status == SyncFileItem::Conflict;
     }
@@ -183,21 +227,40 @@ public:
     /**
      * Whether this item should appear on the protocol tab.
      */
-    bool showInProtocolTab() const
+    [[nodiscard]] bool showInProtocolTab() const
     {
-        return !showInIssuesTab()
+        return (!showInIssuesTab() || _status == SyncFileItem::Restoration)
             // Don't show conflicts that were resolved as "not a conflict after all"
             && !(_instruction == CSYNC_INSTRUCTION_CONFLICT && _status == SyncFileItem::Success);
     }
 
+    [[nodiscard]] bool isEncrypted() const { return _e2eEncryptionStatus != EncryptionStatus::NotEncrypted; }
+
     // Variables useful for everybody
+
+    /** The syncfolder-relative filesystem path that the operation is about
+     *
+     * For rename operation this is the rename source and the target is in _renameTarget.
+     */
     QString _file;
+
+    /** for renames: the name _file should be renamed to
+     * for dehydrations: the name _file should become after dehydration (like adding a suffix)
+     * otherwise empty. Use destination() to find the sync target.
+     */
     QString _renameTarget;
 
+    /** The db-path of this item.
+     *
+     * This can easily differ from _file and _renameTarget if parts of the path were renamed.
+     */
+    QString _originalFile;
+
     /// Whether there's end to end encryption on this file.
     /// If the file is encrypted, the _encryptedFilename is
     /// the encrypted name on the server.
     QString _encryptedFileName;
+
     ItemType _type BITFIELD(3);
     Direction _direction BITFIELD(3);
     bool _serverHasIgnoredFiles BITFIELD(1);
@@ -217,19 +280,21 @@ public:
     // Variables useful to report to the user
     Status _status BITFIELD(4);
     bool _isRestoration BITFIELD(1); // The original operation was forbidden, and this is a restoration
+    bool _isSelectiveSync BITFIELD(1); // The file is removed or ignored because it is in the selective sync list
+    EncryptionStatus _e2eEncryptionStatus = EncryptionStatus::NotEncrypted; // The file is E2EE or the content of the directory should be E2EE
     quint16 _httpErrorCode = 0;
     RemotePermissions _remotePerm;
     QString _errorString; // Contains a string only in case of error
     QByteArray _responseTimeStamp;
+    QByteArray _requestId; // X-Request-Id of the failed request
     quint32 _affectedItems = 1; // the number of affected items by the operation on this item.
     // usually this value is 1, but for removes on dirs, it might be much higher.
 
     // Variables used by the propagator
-    csync_instructions_e _instruction = CSYNC_INSTRUCTION_NONE;
-    QString _originalFile; // as it is in the csync tree
+    SyncInstructions _instruction = CSYNC_INSTRUCTION_NONE;
     time_t _modtime = 0;
     QByteArray _etag;
-    quint64 _size = 0;
+    qint64 _size = 0;
     quint64 _inode = 0;
     QByteArray _fileId;
 
@@ -242,11 +307,28 @@ public:
     QByteArray _checksumHeader;
 
     // The size and modtime of the file getting overwritten (on the disk for downloads, on the server for uploads).
-    quint64 _previousSize = 0;
+    qint64 _previousSize = 0;
     time_t _previousModtime = 0;
 
     QString _directDownloadUrl;
     QString _directDownloadCookies;
+
+    LockStatus _locked = LockStatus::UnlockedItem;
+    QString _lockOwnerId;
+    QString _lockOwnerDisplayName;
+    LockOwnerType _lockOwnerType = LockOwnerType::UserLock;
+    QString _lockEditorApp;
+    qint64 _lockTime = 0;
+    qint64 _lockTimeout = 0;
+
+    bool _isShared = false;
+    time_t _lastShareStateFetchedTimestamp = 0;
+
+    bool _sharedByMe = false;
+
+    bool _isFileDropDetected = false;
+
+    bool _isEncryptedMetadataNeedUpdate = false;
 };
 
 inline bool operator<(const SyncFileItemPtr &item1, const SyncFileItemPtr &item2)
@@ -254,7 +336,7 @@ inline bool operator<(const SyncFileItemPtr &item1, const SyncFileItemPtr &item2
     return *item1 < *item2;
 }
 
-typedef QVector<SyncFileItemPtr> SyncFileItemVector;
+using SyncFileItemVector = QVector<SyncFileItemPtr>;
 }
 
 Q_DECLARE_METATYPE(OCC::SyncFileItem)
